import chess.pgn
from types import SimpleNamespace

import bunseki.util
import bunseki.util as util
import bunseki.asklichess as ali
from collections import defaultdict 
import argparse



parser = argparse.ArgumentParser() 
parser.add_argument('-w','--white',dest='COLOR', help= 'is the repertoire for white', type=int, choices=[0,1], required=True)
parser.add_argument('-i','--pgnfile',dest='PGNFILE', help= 'path to pgn file to parse', type=str, required=True)
parser.add_argument('-u','--utilitycut',dest='UTILITYCUT', help= 'cutoff based on utility..i.e. probability that the top move is on board', type=float, default = 2)
parser.add_argument('-c','--mydb',dest='SHAREDB', help= 'our db is the masterdb, if this is 1 it is the same as the default db', type=int, default = 0)
parser.add_argument('-d','--database',dest='DATABASE', help= '0 = master database 1 = lichess', type=int, default = 0)
parser.add_argument('-s','--lichess_strength',dest='STRENGTH', help= 'select playing strength for lichess db', type=int,nargs='+', default = [1600,1800])
parser.add_argument('-t','--lichess_format',dest='TIMECTL', help= 'select time control for lichess db', type=str,nargs='+', default = ['blitz','rapid'])
parser.add_argument('--min-ply',dest='MINPLY', help= 'dont show alternatives for the first plies', type=int, default = -1)

parser.add_argument('-m','--choose_most_played',dest='MYMOVE',default = 'best', help= 'criterion for choosing our move most_played,terminate, or best', type=str)

def main(): 
    args = parser.parse_args()
    masterarg= SimpleNamespace(DATABASE=0)
    game = util.loadpgn(args.PGNFILE) 
    game.proba = 1
    fens = {} # visited

    db = ali.lichess(args)
    if args.DATABASE == 1 and args.SHAREDB == 0:
        mydb = ali.lichess(masterarg)
    else:
        mydb = db


    '''
    fen = game.board().fen()
    moves, total_games, openingname = h.call( lambda: ali.ask(fen,args),fen)
    game.proba = 1
    game.openingname = openingname
    if args.COLOR ==0:
        proba_calculation(game, moves, total_games)
    '''

    def myturn(gn):
        return gn.ply()% 2 != args.COLOR


    def visit(gn):

        fen = gn.board().fen()
        try:
            moves, movesum, opening = (mydb if myturn(gn) else db).ask(fen)
        except:
            db.end()
            mydb.end()
            assert False, 'sending requests too fast probably'


        fens[fen]=1
        
        #print()
        #print(gn.ply())
        print(gn.board().unicode())
        print ("proba:",gn.__dict__.get("proba","no proba"))
        if gn.ply() < args.MINPLY:
            if myturn(gn):
                for child in gn.variations:
                    child.proba = gn.proba
            else:
                for move in moves:
                    mov = gn.board().push_san(move['san'])
                    if gn.has_variation(mov):
                        gn.variation(mov).proba =  (bunseki.util.sumdi(move) / movesum) * gn.proba

        elif myturn(gn):
            if not gn.variations:
                best_san = util.find_best(moves, gn.ply())

                if args.MYMOVE=='most_frequent':
                    mov = gn.board().push_san(moves[0]['san'])
                    child = gn.add_variation(mov)
                    if best_san != moves[0]['san']:
                        child.comment = f"better: {best_san}"


                elif args.MYMOVE=='best':
                    mov = gn.board().push_san(best_san)
                    child = gn.add_variation(mov)
                    if best_san != moves[0]['san']:
                        child.comment = f"most frequent: {moves[0]['san']}"


                elif args.MYMOVE == 'terminate':
                    if best_san == moves[0]['san']:
                        mov = gn.board().push_san(best_san)
                        child = gn.add_variation(mov)
                    else:
                        gn.comment = f"terminated:{gn.proba*100:.1f}"

                else:
                    assert False

            for child in gn.variations:
                child.proba = gn.proba
        else:
            for move in moves: 
                p= (bunseki.util.sumdi(move) / movesum) * gn.proba
                mov = gn.board().push_san(move['san']) 
                print( '\t',move['san'],int(p*100))
                if p > (args.UTILITYCUT/100):
                    if not  gn.has_variation(mov):
                        gn.add_variation(mov)
                        print("  addvariation", move['san'])
                if gn.has_variation(mov):
                    gn.variation(mov).proba = p


        return gn.variations

    nodelist =  [game]

    while nodelist: 
        node = nodelist.pop()
        if node.board().fen() in fens:
            continue
        else:
            nodelist+=visit(node)


    db.end()
    mydb.end()
    print(game)


if __name__ == '__main__':
    main()



# based on https://github.com/permutationlock/merge-pgn
import chess.pgn
import sys
import chess
import os 
import pickle
def merge(games):
    master_node = chess.pgn.Game()

    mlist = []
    for game in games:
        mlist.extend(game.variations)

    variations = [(master_node, mlist)]
    done = False

    while not done:
        newvars = []
        done = True
        for vnode, nodes in variations:
            newmoves = {}
            for node in nodes:
                if node.move is None:
                    continue
                elif node.move not in list(newmoves):
                    nvnode = vnode.add_variation(node.move)
                    if len(node.variations) > 0:
                        done = False
                    newvars.append((nvnode, node.variations))
                    newmoves[node.move] = len(newvars) - 1
                else:
                    nvnode, nlist = newvars[newmoves[node.move]]
                    if len(node.variations) > 0:
                        done = False
                    nlist.extend(node.variations)
                    newvars[newmoves[node.move]] = (nvnode, nlist)
        variations = newvars

    return master_node

def loadpgn(filename, maxread=999):
    assert maxread == 999, 'not implemented'
    repo = open(filename)
    games = []
    while True:
        stuff = chess.pgn.read_game(repo)
        if not stuff:
            break
        games.append(stuff)
    game = merge(games)
    return game


def getimportance(node):
    res = 1
    while node.parent:
        res*= node.proba
        node = node.parent
    return res


def history(node):
    r=[]
    while node.parent:
        r.append(str(node.move))
        node = node.parent
    r.reverse()
    return r

def pgn(node):
    g = chess.pgn.Game()
    for move in history(node):
        g = g.add_variation(chess.Move.from_uci(move))
    exporter = chess.pgn.StringExporter(headers=False, variations=True, comments=True)
    return g.game().accept(exporter)



class cacher():
    def __init__(self,cachename): 
        self.cachename = f".{cachename}"
        if os.path.exists(self.cachename): 
            self.cache = pickle.load( open( self.cachename, "rb" ) )
        else:
            self.cache={}

    def write(self):
        pickle.dump( self.cache, open( self.cachename, "wb" ) )

    def call(self,f, key): 
        if key in self.cache:
            return self.cache[key]
        else:
            r = f()
            self.cache[key] = r 
        return r


'''
def find_lemons(moves, ply):
    # checks if the most popular move is not the best in win percentage
    # (compared to a move at least 20% played)

    us = 'black'
    them = 'white'

    if (ply+1) % 2:
        us, them = them,us
    score_mv = lambda m: (m[us]-m[them]) / sumdi(m)
    score = score_mv(moves[0])
    sum = max(sumdi(moves[0]), 100)
    if Z:=[ e for e in moves[1:] if sumdi(e) > sum*.2  and score_mv(e) > score  ]:
        return "there might be a better move"+str([ z['san'] for z in Z])
    return ''


def find_best(moves, ply):
    # returns the best move, most played or ''

    us = 'black'
    them = 'white'
    if (ply + 1) % 2:
        us, them = them, us
    sum = max(sumdi(moves[0]), 100)
    score_mv = lambda m: (m[us] - m[them]) / sumdi(m) if sumdi(m) > .2*sum else -99999
    best = max(moves, key=score_mv)
    return best['san'], '' if best['san'] == moves[0]['san'] else f"most played is {moves[0]['san']}"
'''

def find_best(moves, ply):
    # returns the best move, most played or ''

    us = 'black'
    them = 'white'
    if (ply + 1) % 2:
        us, them = them, us
    sum = max(sumdi(moves[0]), 100)
    score_mv = lambda m: (m[us] - m[them]) / sumdi(m) if sumdi(m) > .2*sum else -99999
    best = max(moves, key=score_mv)
    return best['san']


def sumdi(di):
    return di['black']+di['white']+di['draws']


# we interact with this: https://github.com/niklasf/lila-openingexplorer



import requests as req
import time
import bunseki.util as util
from bunseki.util import sumdi

session = req.Session()


class lichess:
    def __init__(self,args):
        self.args=args
        if args.DATABASE == 1:
            hashname = hash((args.DATABASE, tuple(args.STRENGTH), tuple( args.TIMECTL)))
        else:
            hashname = 23
        self.hasher = util.cacher( hashname )
    
    def end(self):
        self.hasher.write()

    def ask(self, fen):
        reply =  self.hasher.call( lambda: self.lookup(fen),fen)
        return reply


    def lookup(self,fen):
        print(".",end='')
        if self.args.DATABASE ==1:
            parm = {'fen':fen, 
                    'topGames': 0,
                    'recentGames':0, 
                    'moves':'15',
                    'variant':"standard",
                    'speeds[]':self.args.TIMECTL,
                    'ratings[]':self.args.STRENGTH}
            res = session.get('https://explorer.lichess.ovh/lichess', params=parm)
        else:
            parm = {'fen':fen, 'topGames': 0, 'moves':'15' }
            res = session.get('https://explorer.lichess.ovh/master', params=parm)

        if  res.status_code == 200:
            js = res.json()
            time.sleep(.8)
            return js['moves'], sumdi(js), js['opening'] or ''
        else:
            print (res.status_code)
            print(res.text)
            assert False


def fmt_stst(my,al):
    return int((my/al)*100)
def fmt_q(di):
    a= sumdi(di)
    i = lambda x: int(x*100)
    return [ i(di[v]/a) for v in['white','draws','black']   ]


def analyze_single(di, mvsum, okmoves):
    # returns percentage, inlist, and display string
    mvcnt = sumdi(di)
    perc = mvcnt/ mvsum
    inlist = di['san'] in okmoves
    description  = f"\t{di['san']}\t{mvcnt}({fmt_stst(mvcnt,mvsum)})\t{'OK' if inlist else '!!'} {fmt_q(di)}"

    return perc, inlist, description, mvcnt


def missvalue(z):
    return max(z)
    if len(z) == 0:
        return 0.01
    return sum(z)/len(z)
    #return sum([i**2 for i in z])/len(z)/1000

def analyse2(move_dict_list,okmoves,minmov, target_share, minperc):

    '''
    move_dict_list: dicts from lichess db
    okmoves: san of the moves we have in our list
    minmov: ignore moves played less than this
    target_share: we want this much coverage
    minperc: move occurange in db
    '''

    sum_moves = sum([sumdi(di) for di in move_dict_list])

    # for every move -> calculate percentage covered and the indicator string
    all_moves  = [ analyze_single (di,sum_moves,okmoves) for di in move_dict_list ]
    perc_ok =  sum([a[0] for a in all_moves if a[1]])
    if perc_ok > target_share:
        perc_missing = 0
    else:

        # was wollen wir hier? ich will messen wie schlecht wir sind...
        # 5+5+5+5 < 20 ... 
        vls = [ a[0] for a in all_moves if a[0]>minperc and not a[1]]
        perc_missing = missvalue([ a[0] for a in all_moves if  not a[1]])
        #print(vls, perc_missing)
        
        #print(perc_missing, perc_ok, sum([a[0] for a in all_moves if a[0] < 5 and not a[1]]))



    ret = '\n'.join([a[2] for a in all_moves if a[3] > minmov and a[0] > minperc])
    return ret, perc_missing 

